@@ -10,6 +10,7 @@
 #include <getopt.h>
 #include <unistd.h>
 #include <errno.h>
+#include <assert.h>
 #include <xcb/xcb.h>
 #include <xcb/xcbext.h>
 #if WITH_XINERAMA
@@ -32,18 +33,19 @@ typedef struct font_t {
 } font_t;
 
 typedef struct monitor_t {
-    int x, y, width;
+    char *name;
+    int x, y, width, height;
     xcb_window_t window;
     xcb_pixmap_t pixmap;
     struct monitor_t *prev, *next;
 } monitor_t;
 
 typedef struct area_t {
-    unsigned int begin:16;
-    unsigned int end:16;
-    bool active:1;
-    int align:3;
-    unsigned int button:3;
+    unsigned begin;
+    unsigned end;
+    bool complete;
+    unsigned align;
+    unsigned button;
     xcb_window_t window;
     char *cmd;
 } area_t;
@@ -59,8 +61,8 @@ typedef union rgba_t {
 } rgba_t;
 
 typedef struct area_stack_t {
-    int at, max;
-    area_t *area;
+    area_t *ptr;
+    unsigned int index, alloc;
 } area_stack_t;
 
 enum {
@@ -81,15 +83,13 @@ enum {
     GC_MAX
 };
 
-#define MAX_FONT_COUNT 5
-
 static xcb_connection_t *c;
 static xcb_screen_t *scr;
 static xcb_gcontext_t gc[GC_MAX];
 static xcb_visualid_t visual;
 static xcb_colormap_t colormap;
 static monitor_t *monhead, *montail;
-static font_t *font_list[MAX_FONT_COUNT];
+static font_t **font_list = NULL;
 static int font_count = 0;
 static int font_index = -1;
 static uint32_t attrs = 0;
@@ -101,6 +101,12 @@ static rgba_t fgc, bgc, ugc;
 static rgba_t dfgc, dbgc, dugc;
 static area_stack_t area_stack;
 
+static const rgba_t BLACK = (rgba_t){ .r = 0, .g = 0, .b = 0, .a = 255 };
+static const rgba_t WHITE = (rgba_t){ .r = 255, .g = 255, .b = 255, .a = 255 };
+
+static int num_outputs = 0;
+static char **output_names = NULL;
+
 void
 update_gc (void)
 {
@@ -347,8 +353,8 @@ area_t *
 area_get (xcb_window_t win, const int btn, const int x)
 {
     // Looping backwards ensures that we get the innermost area first
-    for (int i = area_stack.at - 1; i >= 0; i--) {
-        area_t *a = &area_stack.area[i];
+    for (int i = area_stack.index - 1; i >= 0; i--) {
+        area_t *a = &area_stack.ptr[i];
         if (a->window == win && a->button == btn && x >= a->begin && x < a->end)
             return a;
     }
@@ -363,9 +369,9 @@ area_shift (xcb_window_t win, const int align, int delta)
     if (align == ALIGN_C)
         delta /= 2;
 
-    for (int i = 0; i < area_stack.at; i++) {
-        area_t *a = &area_stack.area[i];
-        if (a->window == win && a->align == align && !a->active) {
+    for (int i = 0; i < area_stack.index; i++) {
+        area_t *a = &area_stack.ptr[i];
+        if (a->window == win && a->align == align && !a->complete) {
             a->begin -= delta;
             a->end -= delta;
         }
@@ -384,9 +390,9 @@ area_add (char *str, const char *optend, char **end, monitor_t *mon, const int x
         *end = str;
 
         // Find most recent unclosed area.
-        for (i = area_stack.at - 1; i >= 0 && !area_stack.area[i].active; i--)
+        for (i = area_stack.index - 1; i >= 0 && !area_stack.ptr[i].complete; i--)
             ;
-        a = &area_stack.area[i];
+        a = &area_stack.ptr[i];
 
         // Basic safety checks
         if (!a->cmd || a->align != align || a->window != mon->window) {
@@ -411,16 +417,19 @@ area_add (char *str, const char *optend, char **end, monitor_t *mon, const int x
                 break;
         }
 
-        a->active = false;
+        a->complete = false;
         return true;
     }
 
-    if (area_stack.at + 1 > area_stack.max) {
-        fprintf(stderr, "Cannot add any more clickable areas (used %d/%d)\n", 
-                area_stack.at, area_stack.max);
-        return false;
+    if (area_stack.index + 1 > area_stack.alloc) {
+        area_stack.ptr = realloc(area_stack.ptr, sizeof(area_t) * (area_stack.index + 1));
+        if (!area_stack.ptr) {
+            fprintf(stderr, "Failed to allocate new input areas\n");
+            exit(EXIT_FAILURE);
+        }
+        area_stack.alloc += 1;
     }
-    a = &area_stack.area[area_stack.at++];
+    a = &area_stack.ptr[area_stack.index++];
 
     // Found the closing : and check if it's just an escaped one
     for (trail = strchr(++str, ':'); trail && trail[-1] == '\\'; trail = strchr(trail + 1, ':'))
@@ -445,7 +454,7 @@ area_add (char *str, const char *optend, char **end, monitor_t *mon, const int x
 
     // This is a pointer to the string buffer allocated in the main
     a->cmd = str;
-    a->active = true;
+    a->complete = true;
     a->align = align;
     a->begin = x;
     a->window = mon->window;
@@ -485,6 +494,18 @@ select_drawable_font (const uint16_t c)
     return NULL;
 }
 
+int
+pos_to_absolute(monitor_t *mon, int pos, int align)
+{
+    switch (align) {
+        case ALIGN_L: return pos;
+        case ALIGN_R: return mon->width - pos;
+        case ALIGN_C: return mon->width / 2 + pos / 2;
+    }
+
+    return 0;
+}
+
 void
 parse (char *text)
 {
@@ -492,14 +513,21 @@ parse (char *text)
     monitor_t *cur_mon;
     int pos_x, align, button;
     char *p = text, *block_end, *ep;
-    rgba_t tmp;
 
     pos_x = 0;
     align = ALIGN_L;
     cur_mon = monhead;
 
+    // Reset the default color set
+    bgc = dbgc;
+    fgc = dfgc;
+    ugc = fgc;
+    update_gc();
+    // Reset the default attributes
+    attrs = 0;
+
     // Reset the stack position
-    area_stack.at = 0;
+    area_stack.index = 0;
 
     for (monitor_t *m = monhead; m != NULL; m = m->next)
         fill_rect(m->pixmap, gc[GC_CLEAR], 0, 0, m->width, bh);
@@ -511,88 +539,152 @@ parse (char *text)
         if (p[0] == '%' && p[1] == '{' && (block_end = strchr(p++, '}'))) {
             p++;
             while (p < block_end) {
-                int w;
                 while (isspace(*p))
                     p++;
 
                 switch (*p++) {
+                    // Enable/disable attributes.
                     case '+': set_attribute('+', *p++); break;
                     case '-': set_attribute('-', *p++); break;
                     case '!': set_attribute('!', *p++); break;
 
-                    case 'R':
-                              tmp = fgc;
-                              fgc = bgc;
-                              bgc = tmp;
-                              update_gc();
-                              break;
+                    // Reverse foreground/background color.
+                    case 'R': {
+                        rgba_t tmp = fgc;
+                        fgc = bgc;
+                        bgc = tmp;
+                        update_gc();
+                    } break;
 
-                    case 'l': pos_x = 0; align = ALIGN_L; break;
-                    case 'c': pos_x = 0; align = ALIGN_C; break;
-                    case 'r': pos_x = 0; align = ALIGN_R; break;
+                    // Alignment specifiers.
+                    // Keep track of where we are and where we're moving to so
+                    // that underlines/overlines are correctly drawn over the
+                    // empty space.
+                    case 'l': {
+                        int left_ep = 0;
+                        int right_ep = pos_to_absolute(cur_mon, pos_x, align);
+                        draw_lines(cur_mon, left_ep, right_ep - left_ep);
+                        pos_x = 0; align = ALIGN_L;
+                    } break;
+                    case 'c': {
+                        int left_ep = pos_to_absolute(cur_mon, pos_x, align);
+                        int right_ep = cur_mon->width / 2;
+                        if (right_ep < left_ep) {
+                            int tmp = left_ep;
+                            left_ep = right_ep;
+                            right_ep = tmp;
+                        }
+                        draw_lines(cur_mon, left_ep, right_ep - left_ep);
+                        pos_x = 0; align = ALIGN_C;
+                    } break;
+                    case 'r': {
+                        int left_ep = pos_to_absolute(cur_mon, pos_x, align);
+                        int right_ep = cur_mon->width;
+                        if (right_ep < left_ep) {
+                            int tmp = left_ep;
+                            left_ep = right_ep;
+                            right_ep = tmp;
+                        }
+                        draw_lines(cur_mon, left_ep, right_ep - left_ep);
+                        pos_x = 0; align = ALIGN_R;
+                    } break;
 
-                    case 'A':
-                              button = XCB_BUTTON_INDEX_1;
-                              // The range is 1-5
-                              if (isdigit(*p) && (*p > '0' && *p < '6'))
-                                  button = *p++ - '0';
-                              if (!area_add(p, block_end, &p, cur_mon, pos_x, align, button))
-                                  return;
-                              break;
+                    // Define input area.
+                    case 'A': {
+                        button = XCB_BUTTON_INDEX_1;
+                        // The range is 1-5
+                        if (isdigit(*p) && (*p > '0' && *p < '6'))
+                            button = *p++ - '0';
+                        if (!area_add(p, block_end, &p, cur_mon, pos_x, align, button))
+                            return;
+                    } break;
 
+                    // Set background/foreground/underline color.
                     case 'B': bgc = parse_color(p, &p, dbgc); update_gc(); break;
                     case 'F': fgc = parse_color(p, &p, dfgc); update_gc(); break;
                     case 'U': ugc = parse_color(p, &p, dugc); update_gc(); break;
 
-                    case 'S':
-                              if (*p == '+' && cur_mon->next)
-                              { cur_mon = cur_mon->next; }
-                              else if (*p == '-' && cur_mon->prev)
-                              { cur_mon = cur_mon->prev; }
-                              else if (*p == 'f')
-                              { cur_mon = monhead; }
-                              else if (*p == 'l')
-                              { cur_mon = montail ? montail : monhead; }
-                              else if (isdigit(*p))
-                              { cur_mon = monhead;
+                    // Set current monitor used for drawing.
+                    case 'S': {
+                        monitor_t *orig_mon = cur_mon;
+
+                        switch (*p) {
+                            case '+': // Next monitor.
+                                if (cur_mon->next) cur_mon = cur_mon->next;
+                                p += 1;
+                                break;
+                            case '-': // Previous monitor.
+                                if (cur_mon->prev) cur_mon = cur_mon->prev;
+                                p += 1;
+                                break;
+                            case 'f': // First monitor.
+                                cur_mon = monhead;
+                                p += 1;
+                                break;
+                            case 'l': // Last monitor.
+                                cur_mon = montail ? montail : monhead;
+                                p += 1;
+                                break;
+                            case 'n': { // Named monitor.
+                                const size_t name_len = block_end - (p + 1);
+                                cur_mon = monhead;
+                                while (cur_mon->next) {
+                                    if (cur_mon->name &&
+                                            !strncmp(cur_mon->name, p + 1, name_len))
+                                        break;
+                                    cur_mon = cur_mon->next;
+                                }
+                                p += 1 + name_len;
+                            } break;
+                            case '0' ... '9': // Numbered monitor.
+                                cur_mon = monhead;
                                 for (int i = 0; i != *p-'0' && cur_mon->next; i++)
                                     cur_mon = cur_mon->next;
-                              }
-                              else
-                              { p++; continue; }
+                                p += 1;
+                                break;
+                            default:
+                                fprintf(stderr, "Unknown S specifier '%c'\n", *p++);
+                                break;
+                        }
 
+                        if (orig_mon != cur_mon) {
+                            pos_x = 0;
+                            align = ALIGN_L;
+                        }
+                    } break;
+
+                    // Draw a N-pixel wide empty character.
+                    case 'O': {
+                        errno = 0;
+                        int w = (int) strtoul(p, &p, 10);
+                        if (errno)
+                            continue;
+
+                        draw_shift(cur_mon, pos_x, align, w);
+
+                        pos_x += w;
+                        area_shift(cur_mon->window, align, w);
+                    } break;
+
+                    case 'T': {
+                          if (*p == '-') {
+                              // Switch to automatic font selection.
+                              font_index = -1;
                               p++;
-                              pos_x = 0;
-                              break;
-                    case 'O':
-                              errno = 0;
-                              w = (int) strtoul(p, &p, 10);
-                              if (errno)
-                                  continue;
-
-                              draw_shift(cur_mon, pos_x, align, w);
-
-                              pos_x += w;
-                              area_shift(cur_mon->window, align, w);
-                              break;
-
-                    case 'T':
-                              if (*p == '-') { //Reset to automatic font selection
+                          } else if (isdigit(*p)) {
+                              font_index = (int)strtoul(p, &ep, 10);
+                              // User-specified 'font_index' ∊ (0,font_count]
+                              // Otherwise just fallback to the automatic font selection
+                              if (!font_index || font_index > font_count) {
+                                  fprintf(stderr, "Invalid font index %d\n", font_index);
                                   font_index = -1;
-                                  p++;
-                                  break;
-                              } else if (isdigit(*p)) {
-                                  font_index = (int)strtoul(p, &ep, 10);
-                                  // User-specified 'font_index' ∊ (0,font_count]
-                                  // Otherwise just fallback to the automatic font selection
-                                  if (!font_index || font_index > font_count)
-                                  font_index = -1;
-                                  p = ep;
-                                  break;
-                              } else {
-                                  fprintf(stderr, "Invalid font slot \"%c\"\n", *p++); //Swallow the token
-                                  break;
                               }
+                              p = ep;
+                          } else {
+                              // Swallow the invalid character and keep parsing.
+                              fprintf(stderr, "Invalid font slot \"%c\"\n", *p++);
+                          }
+                    } break;
 
                     // In case of error keep parsing after the closing }
                     default:
@@ -662,11 +754,6 @@ parse (char *text)
 void
 font_load (const char *pattern)
 {
-    if (font_count >= MAX_FONT_COUNT) {
-        fprintf(stderr, "Max font count reached. Could not load font \"%s\"\n", pattern);
-        return;
-    }
-
     xcb_query_font_cookie_t queryreq;
     xcb_query_font_reply_t *font_info;
     xcb_void_cookie_t cookie;
@@ -681,9 +768,10 @@ font_load (const char *pattern)
     }
 
     font_t *ret = calloc(1, sizeof(font_t));
-
-    if (!ret)
-        return;
+    if (!ret) {
+        fprintf(stderr, "Failed to allocate new font descriptor\n");
+        exit(EXIT_FAILURE);
+    }
 
     queryreq = xcb_query_font(c, font);
     font_info = xcb_query_font_reply(c, queryreq, NULL);
@@ -704,6 +792,11 @@ font_load (const char *pattern)
 
     free(font_info);
 
+    font_list = realloc(font_list, sizeof(font_t) * (font_count + 1));
+    if (!font_list) {
+        fprintf(stderr, "Failed to allocate %d font descriptors", font_count + 1);
+        exit(EXIT_FAILURE);
+    }
     font_list[font_count++] = ret;
 }
 
@@ -773,7 +866,7 @@ set_ewmh_atoms (void)
 }
 
 monitor_t *
-monitor_new (int x, int y, int width, int height)
+monitor_new (int x, int y, int width, int height, char *name)
 {
     monitor_t *ret;
 
@@ -783,6 +876,7 @@ monitor_new (int x, int y, int width, int height)
         exit(EXIT_FAILURE);
     }
 
+    ret->name = name;
     ret->x = x;
     ret->y = (topbar ? by : height - bh - by) + y;
     ret->width = width;
@@ -819,38 +913,34 @@ monitor_add (monitor_t *mon)
 }
 
 int
-rect_sort_cb (const void *p1, const void *p2)
+mon_sort_cb (const void *p1, const void *p2)
 {
-    const xcb_rectangle_t *r1 = (xcb_rectangle_t *)p1;
-    const xcb_rectangle_t *r2 = (xcb_rectangle_t *)p2;
+    const monitor_t *m1 = (monitor_t *)p1;
+    const monitor_t *m2 = (monitor_t *)p2;
 
-    if (r1->x < r2->x || r1->y + r1->height <= r2->y)
-    {
+    if (m1->x < m2->x || m1->y + m1->height <= m2->y)
         return -1;
-    }
-
-    if (r1->x > r2->x || r1->y + r1->height > r2->y)
-    {
-        return 1;
-    }
+    if (m1->x > m2->x || m1->y + m1->height > m2->y)
+        return  1;
 
     return 0;
 }
 
 void
-monitor_create_chain (xcb_rectangle_t *rects, const int num)
+monitor_create_chain (monitor_t *mons, const int num)
 {
     int i;
     int width = 0, height = 0;
     int left = bx;
 
-    // Sort before use
-    qsort(rects, num, sizeof(xcb_rectangle_t), rect_sort_cb);
+    // Sort before use, but only if specific outputs were not specified on command line
+    if (!num_outputs)
+        qsort(mons, num, sizeof(monitor_t), mon_sort_cb);
 
     for (i = 0; i < num; i++) {
-        int h = rects[i].y + rects[i].height;
+        int h = mons[i].y + mons[i].height;
         // Accumulated width of all monitors
-        width += rects[i].width;
+        width += mons[i].width;
         // Get height of screen from y_offset + height of lowest monitor
         if (h >= height)
         height = h;
@@ -872,28 +962,29 @@ monitor_create_chain (xcb_rectangle_t *rects, const int num)
     // Left is a positive number or zero therefore monitors with zero width are excluded
     width = bw;
     for (i = 0; i < num; i++) {
-        if (rects[i].y + rects[i].height < by)
+        if (mons[i].y + mons[i].height < by)
             continue;
-        if (rects[i].width > left) {
+        if (mons[i].width > left) {
             monitor_t *mon = monitor_new(
-                    rects[i].x + left,
-                    rects[i].y,
-                    min(width, rects[i].width - left),
-                    rects[i].height);
+                    mons[i].x + left,
+                    mons[i].y,
+                    min(width, mons[i].width - left),
+                    mons[i].height,
+                    mons[i].name? strdup(mons[i].name) : NULL);
 
             if (!mon)
                 break;
 
             monitor_add(mon);
 
-            width -= rects[i].width - left;
+            width -= mons[i].width - left;
 
             // No need to check for other monitors
             if (width <= 0)
                 break;
         }
 
-        left -= rects[i].width;
+        left -= mons[i].width;
 
         if (left < 0)
             left = 0;
@@ -918,14 +1009,19 @@ get_randr_monitors (void)
     num = xcb_randr_get_screen_resources_current_outputs_length(rres_reply);
     outputs = xcb_randr_get_screen_resources_current_outputs(rres_reply);
 
-
     // There should be at least one output
     if (num < 1) {
         free(rres_reply);
         return;
     }
 
-    xcb_rectangle_t rects[num];
+    // Every entry starts with a size of 0, making it invalid until we fill in
+    // the data retrieved from the Xserver.
+    monitor_t *mons = calloc(max(num, num_outputs), sizeof(monitor_t));
+    if (!mons) {
+        fprintf(stderr, "failed to allocate the monitor array\n");
+        return;
+    }
 
     // Get all outputs
     for (i = 0; i < num; i++) {
@@ -937,61 +1033,96 @@ get_randr_monitors (void)
         // Output disconnected or not attached to any CRTC ?
         if (!oi_reply || oi_reply->crtc == XCB_NONE || oi_reply->connection != XCB_RANDR_CONNECTION_CONNECTED) {
             free(oi_reply);
-            rects[i].width = 0;
             continue;
         }
 
         ci_reply = xcb_randr_get_crtc_info_reply(c,
                 xcb_randr_get_crtc_info(c, oi_reply->crtc, XCB_CURRENT_TIME), NULL);
 
-        free(oi_reply);
-
         if (!ci_reply) {
-            fprintf(stderr, "Failed to get RandR ctrc info\n");
+            fprintf(stderr, "Failed to get RandR crtc info\n");
             free(rres_reply);
-            return;
+            goto cleanup_mons;
         }
 
-        // There's no need to handle rotated screens here (see #69)
-        rects[i] = (xcb_rectangle_t){ ci_reply->x, ci_reply->y, ci_reply->width, ci_reply->height };
+        int name_len = xcb_randr_get_output_info_name_length(oi_reply);
+        uint8_t *name_ptr = xcb_randr_get_output_info_name(oi_reply);
 
+        bool is_valid = true;
+
+        if (num_outputs) {
+            // Skip outputs missing from the list.
+            is_valid = false;
+            // Allocate monitors following the specified order.
+            for (j = 0; j < num_outputs; j++) {
+                // Already allocated, the list contains a duplicate.
+                if (mons[j].name)
+                    break;
+
+                if (!memcmp(output_names[j], name_ptr, name_len) &&
+                        strlen(output_names[j]) == name_len) {
+                    is_valid = true;
+                    break;
+                }
+            }
+        }
+
+        if (is_valid) {
+            char *alloc_name = calloc(name_len + 1, 1);
+            if (!alloc_name) {
+                fprintf(stderr, "failed to allocate output name\n");
+                exit(EXIT_FAILURE);
+            }
+            memcpy(alloc_name, name_ptr, name_len);
+
+            // There's no need to handle rotated screens here (see #69)
+            mons[i] = (monitor_t){ alloc_name, ci_reply->x, ci_reply->y,
+                ci_reply->width, ci_reply->height, 0, 0, NULL, NULL };
+            valid += 1;
+        }
+
+        free(oi_reply);
         free(ci_reply);
-
-        valid++;
     }
 
     free(rres_reply);
 
     // Check for clones and inactive outputs
     for (i = 0; i < num; i++) {
-        if (rects[i].width == 0)
+        if (mons[i].width == 0)
             continue;
 
         for (j = 0; j < num; j++) {
             // Does I contain J ?
 
-            if (i != j && rects[j].width) {
-                if (rects[j].x >= rects[i].x && rects[j].x + rects[j].width <= rects[i].x + rects[i].width &&
-                    rects[j].y >= rects[i].y && rects[j].y + rects[j].height <= rects[i].y + rects[i].height) {
-                    rects[j].width = 0;
+            if (i != j && mons[j].width && !mons[j].name) {
+                if (mons[j].x >= mons[i].x && mons[j].x + mons[j].width <= mons[i].x + mons[i].width &&
+                    mons[j].y >= mons[i].y && mons[j].y + mons[j].height <= mons[i].y + mons[i].height) {
+                    mons[j].width = 0;
                     valid--;
                 }
             }
         }
     }
 
-    if (valid < 1) {
+    if (valid > 0) {
+        monitor_t valid_mons[valid];
+        for (i = j = 0; i < num && j < valid; i++) {
+            if (mons[i].width != 0) {
+                valid_mons[j++] = mons[i];
+            }
+        }
+
+        monitor_create_chain(valid_mons, valid);
+    } else {
         fprintf(stderr, "No usable RandR output found\n");
-        return;
     }
 
-    xcb_rectangle_t r[valid];
-
-    for (i = j = 0; i < num && j < valid; i++)
-        if (rects[i].width != 0)
-            r[j++] = rects[i];
-
-    monitor_create_chain(r, valid);
+cleanup_mons:
+    for (i = 0; i < num; i++) {
+        free(mons[i].name);
+    }
+    free(mons);
 }
 
 #ifdef WITH_XINERAMA
@@ -1002,26 +1133,32 @@ get_xinerama_monitors (void)
     xcb_xinerama_screen_info_iterator_t iter;
     int screens;
 
+    if (num_outputs) {
+        fprintf(stderr, "Using output names with Xinerama is not yet supported\n");
+        return;
+    }
+
     xqs_reply = xcb_xinerama_query_screens_reply(c,
             xcb_xinerama_query_screens_unchecked(c), NULL);
 
     iter = xcb_xinerama_query_screens_screen_info_iterator(xqs_reply);
     screens = iter.rem;
 
-    xcb_rectangle_t rects[screens];
+    monitor_t mons[screens];
 
     // Fetch all the screens first
     for (int i = 0; iter.rem; i++) {
-        rects[i].x = iter.data->x_org;
-        rects[i].y = iter.data->y_org;
-        rects[i].width = iter.data->width;
-        rects[i].height = iter.data->height;
+        mons[i].name = NULL;
+        mons[i].x = iter.data->x_org;
+        mons[i].y = iter.data->y_org;
+        mons[i].width = iter.data->width;
+        mons[i].height = iter.data->height;
         xcb_xinerama_screen_info_next(&iter);
     }
 
     free(xqs_reply);
 
-    monitor_create_chain(rects, screens);
+    monitor_create_chain(mons, screens);
 }
 #endif
 
@@ -1097,6 +1234,19 @@ parse_geometry_string (char *str, int *tmp)
     return true;
 }
 
+void
+parse_output_string(char *str)
+{
+    if (!str || !*str)
+        return;
+    output_names = realloc(output_names, sizeof(void *) * (num_outputs + 1));
+    if (!output_names) {
+        fprintf(stderr, "failed to allocate output name\n");
+        exit(EXIT_FAILURE);
+    }
+    output_names[num_outputs++] = strdup(str);
+}
+
 void
 xconn (void)
 {
@@ -1121,7 +1271,7 @@ void
 init (char *wm_name)
 {
     // Try to load a default font
-    if (!font_count)
+    if (font_count == 0)
         font_load("fixed");
 
     // We tried and failed hard, there's something wrong
@@ -1166,6 +1316,11 @@ init (char *wm_name)
     }
 #endif
 
+    if (!monhead && num_outputs != 0) {
+        fprintf(stderr, "Failed to find any specified outputs\n");
+        exit(EXIT_FAILURE);
+    }
+
     if (!monhead) {
         // If I fits I sits
         if (bw < 0)
@@ -1182,7 +1337,7 @@ init (char *wm_name)
         }
 
         // If no RandR outputs or Xinerama screens, fall back to using whole screen
-        monhead = monitor_new(0, 0, bw, scr->height_in_pixels);
+        monhead = monitor_new(0, 0, bw, scr->height_in_pixels, NULL);
     }
 
     if (!monhead)
@@ -1221,18 +1376,25 @@ init (char *wm_name)
 void
 cleanup (void)
 {
-    free(area_stack.area);
+    for (int i = 0; i < num_outputs; i++) {
+        free(output_names[i]);
+    }
+    free(output_names);
+
+    free(area_stack.ptr);
 
     for (int i = 0; i < font_count; i++) {
         xcb_close_font(c, font_list[i]->ptr);
         free(font_list[i]->width_lut);
         free(font_list[i]);
     }
+    free(font_list);
 
     while (monhead) {
         monitor_t *next = monhead->next;
         xcb_destroy_window(c, monhead->window);
         xcb_free_pixmap(c, monhead->pixmap);
+        free(monhead->name);
         free(monhead);
         monhead = next;
     }
@@ -1267,9 +1429,10 @@ main (int argc, char **argv)
     xcb_expose_event_t *expose_ev;
     xcb_button_press_event_t *press_ev;
     char input[4096] = {0, };
+    size_t input_offset = 0;
     bool permanent = false;
     int geom_v[4] = { -1, -1, 0, 0 };
-    int ch, areas;
+    int ch;
     char *wm_name;
 
     // Install the parachute!
@@ -1278,28 +1441,27 @@ main (int argc, char **argv)
     signal(SIGTERM, sighandle);
 
     // B/W combo
-    dbgc = bgc = (rgba_t)0x00000000U;
-    dfgc = fgc = (rgba_t)0xffffffffU;
+    dbgc = bgc = BLACK;
+    dfgc = fgc = WHITE;
     dugc = ugc = fgc;
 
     // A safe default
-    areas = 10;
     wm_name = NULL;
 
     // Connect to the Xserver and initialize scr
     xconn();
 
-    while ((ch = getopt(argc, argv, "hg:bdf:a:pu:B:F:U:n:")) != -1) {
+    while ((ch = getopt(argc, argv, "hg:o:bdf:a:pu:B:F:U:n:")) != -1) {
         switch (ch) {
             case 'h':
                 printf ("lemonbar version %s\n", VERSION);
-                printf ("usage: %s [-h | -g | -b | -d | -f | -a | -p | -n | -u | -B | -F]\n"
+                printf ("usage: %s [-h | -g | -o | -b | -d | -f | -p | -n | -u | -B | -F]\n"
                         "\t-h Show this help\n"
                         "\t-g Set the bar geometry {width}x{height}+{xoffset}+{yoffset}\n"
+                        "\t-o Add randr output by name\n"
                         "\t-b Put the bar at the bottom of the screen\n"
                         "\t-d Force docking (use this if your WM isn't EWMH compliant)\n"
                         "\t-f Set the font name to use\n"
-                        "\t-a Number of clickable areas available (default is 10)\n"
                         "\t-p Don't close after the data ends\n"
                         "\t-n Set the WM_NAME atom to the specified value for this bar\n"
                         "\t-u Set the underline/overline height in pixels\n"
@@ -1307,33 +1469,27 @@ main (int argc, char **argv)
                         "\t-F Set foreground color in #AARRGGBB\n", argv[0]);
                 exit (EXIT_SUCCESS);
             case 'g': (void)parse_geometry_string(optarg, geom_v); break;
+            case 'o': (void)parse_output_string(optarg); break;
             case 'p': permanent = true; break;
             case 'n': wm_name = strdup(optarg); break;
             case 'b': topbar = false; break;
             case 'd': dock = true; break;
             case 'f': font_load(optarg); break;
             case 'u': bu = strtoul(optarg, NULL, 10); break;
-            case 'B': dbgc = bgc = parse_color(optarg, NULL, (rgba_t)0x00000000U); break;
-            case 'F': dfgc = fgc = parse_color(optarg, NULL, (rgba_t)0xffffffffU); break;
+            case 'B': dbgc = bgc = parse_color(optarg, NULL, BLACK); break;
+            case 'F': dfgc = fgc = parse_color(optarg, NULL, WHITE); break;
             case 'U': dugc = ugc = parse_color(optarg, NULL, fgc); break;
-            case 'a': areas = strtoul(optarg, NULL, 10); break;
         }
     }
 
     // Initialize the stack holding the clickable areas
-    area_stack.at = 0;
-    area_stack.max = areas;
-    if (areas) {
-        area_stack.area = calloc(areas, sizeof(area_t));
-
-        if (!area_stack.area) {
-            fprintf(stderr, "Could not allocate enough memory for %d clickable areas, try lowering the number\n", areas);
-            return EXIT_FAILURE;
-        }
+    area_stack.index = 0;
+    area_stack.alloc = 10;
+    area_stack.ptr = calloc(10, sizeof(area_t));
+    if (!area_stack.ptr) {
+        fprintf(stderr, "Failed to allocate enough input areas\n");
+        return EXIT_FAILURE;
     }
-    else
-        area_stack.area = NULL;
-
 
     // Copy the geometry values in place
     bw = geom_v[0];
@@ -1361,11 +1517,45 @@ main (int argc, char **argv)
                 else break;                         // ...bail out
             }
             if (pollin[0].revents & POLLIN) { // New input, process it
-                if (fgets(input, sizeof(input), stdin) == NULL)
-                    break; // EOF received
+                while (true) {
+                    ssize_t r = read(STDIN_FILENO, input + input_offset,
+                            sizeof(input) - input_offset);
+                    if (r == 0) break;
+                    if (r < 0) {
+                        if (errno == EINTR) continue;
+                        exit(EXIT_FAILURE);
+                    }
 
-                parse(input);
-                redraw = true;
+                    input_offset += r;
+
+                    // Try to find the last complete input line in the buffer.
+                    char *input_end = input + input_offset;
+                    char *last_nl = memrchr(input, '\n', input_end - input);
+
+                    if (last_nl) {
+                        char *prev_nl = (last_nl != input) ?
+                                memrchr(input, '\n', last_nl - 1 - input) : NULL;
+                        char *begin = prev_nl? prev_nl + 1: input;
+
+                        *last_nl = '\0';
+
+                        parse(begin);
+                        redraw = true;
+
+                        // Move the unparsed part back to the beginning.
+                        const size_t remaining = input_end - (last_nl + 1);
+                        if (remaining != 0) memmove(input, last_nl + 1, remaining);
+                        input_offset = remaining;
+
+                        break;
+                    }
+
+                    // The input buffer is full and we haven't seen a newline
+                    // yet, discard everything and start from zero.
+                    if (sizeof(input) == input_offset) {
+                        input_offset = 0;
+                    }
+                }
             }
             if (pollin[1].revents & POLLIN) { // The event comes from the Xorg server
                 while ((ev = xcb_poll_for_event(c))) {
